import numpy as np
import astropy.units as u

from .. import SRegion, patch_from_polygon


def test_sregion():
    """
    Test SRegion object
    """
    # From arrays
    x = np.array([0, 0, 1, 1])
    y = np.array([0, 1, 1, 0])
    sr = SRegion(np.array([x, y]).T)

    assert sr.N == 1

    assert np.allclose(sr.centroid[0], 0.5, rtol=1.0e-3)

    assert sr.area[0] == 1.0

    # Converters
    assert hasattr(sr.path[0], "contains_point")
    assert hasattr(sr.shapely[0], "boundary")
    assert hasattr(sr.patch()[0], "get_fc")

    regstr = (
        "polygon(0.000000,0.000000,0.000000,1.000000,"
        + "1.000000,1.000000,1.000000,0.000000)"
    )
    assert sr.region[0] == regstr

    sr.label = "test"
    assert sr.region[0] == regstr + " #  text={test}"

    # SQL string
    pstr = "((0.000,0.000),(0.000,1.000),(1.000,1.000),(1.000,0.000))"
    assert sr.polystr(precision=3)[0] == pstr

    # From s_region string
    for prefix in ["POLYGON", "POLYGON ICRS"]:
        sr.SREGION_PREFIX = prefix
        pstr = (
            f"{prefix} 0.000000 0.000000 0.000000 1.000000 "
            + "1.000000 1.000000 1.000000 0.000000"
        )

        assert pstr == sr.s_region

    snew = SRegion(pstr)
    assert snew.area[0] == 1.0

    # From polygon
    snew = SRegion(sr.shapely[0])
    assert snew.area[0] == 1.0

    # Compound regions
    x2 = np.array([0, 0, 1, 1]) + 2
    y2 = np.array([0, 1, 1, 0]) + 2
    s2 = SRegion(np.array([x2, y2]).T)

    un = sr.union(s2.shapely[0], as_polygon=True)

    assert un.area == 2.0

    # Multiple string
    comp = SRegion(" ".join([sr.s_region, s2.s_region]))
    assert np.allclose(comp.area, 1.0, rtol=1.0e-3)

    un = comp.union(as_polygon=True)
    assert un.area == 2.0

    # Intersects
    assert sr.intersects(s2.shapely[0]) is False

    x3 = np.array([0, 0, 1, 1]) + 0.5
    y3 = np.array([0, 1, 1, 0]) + 0.5
    s3 = SRegion(np.array([x3, y3]).T)

    assert sr.intersects(s3.shapely[0]) is True


def test_wrap():
    """ """
    x = np.array([0, 0, 1, 1]) - 177
    y = np.array([0, 1, 1, 0])
    sr = SRegion(np.array([x, y]).T, wrap=True)

    assert sr.N == 1
    assert np.allclose(sr.area, 1.0)
    assert np.allclose(sr.centroid, [360 - 177 + 0.5, 0.5])


def test_circles():
    """
    Initialize from ``CIRCLE X Y R``
    """
    # CIRCLE string
    circ = SRegion("CIRCLE 5. 5. 1", ncircle=256)
    assert np.allclose(circ.area, np.pi, rtol=1.0e-3)
    assert np.allclose(circ.centroid[0], 5.0, rtol=1.0e-3)

    # R = 2, no units
    circ2 = SRegion("CIRCLE 5. 5. 2", ncircle=256)
    assert np.allclose(circ2.area, 2**2 * np.pi, rtol=1.0e-3)
    assert np.allclose(circ2.centroid[0], 5.0, rtol=1.0e-3)

    # R = 1 arcsec
    cosd = np.cos(45.0 / 180 * np.pi)
    csky = SRegion('CIRCLE 45. 45. 1"', ncircle=256)
    assert np.allclose(csky.area, np.pi / 3600.0**2 / cosd, rtol=1.0e-3)
    assert np.allclose(
        csky.sky_area(unit=u.arcsec**2), np.pi * u.arcsec**2, rtol=1.0e-3
    )

    # R = 1 arcmin
    cosd = np.cos(45.0 / 180 * np.pi)
    csky = SRegion("CIRCLE 45. 45. 1'", ncircle=256)
    assert np.allclose(csky.area, np.pi / 3600.0 / cosd, rtol=1.0e-3)
    assert np.allclose(
        csky.sky_area(unit=u.arcsec**2), 3600 * np.pi * u.arcsec**2, rtol=1.0e-3
    )
    assert np.allclose(csky.centroid[0], 45.0, rtol=1.0e-3)

    # Sky buffer
    cosd = np.cos(45.0 / 180 * np.pi)
    csky = SRegion('CIRCLE 45. 45. 0.001"', ncircle=256)
    csky.sky_buffer(1.0)

    assert np.allclose(csky.area, np.pi / cosd, rtol=1.0e-3)
    assert np.allclose(csky.sky_area(unit=u.deg**2), np.pi * u.deg**2, rtol=1.0e-3)


def test_boxes():
    """
    Initialize from ``BOX X Y W H``
    """
    # BOX string
    box = SRegion("BOX 90 10 5 5")
    assert box.N == 1

    assert np.allclose(box.centroid[0][0], 90, rtol=1.0e-3)
    assert np.allclose(box.centroid[0][1], 10, rtol=1.0e-3)

    assert box.area[0] == 25.0

    # BOX ICRS string
    box = SRegion("BOX ICRS 90 10 5 5")
    assert box.N == 1

    assert np.allclose(box.centroid[0][0], 90, rtol=1.0e-3)
    assert np.allclose(box.centroid[0][1], 10, rtol=1.0e-3)

    assert box.area[0] == 25.0

    # size with units
    box = SRegion("BOX ICRS 90 10 5' 5'")
    assert box.N == 1

    assert np.allclose(box.centroid[0][0], 90, rtol=1.0e-3)
    assert np.allclose(box.centroid[0][1], 10, rtol=1.0e-3)
    assert np.allclose(box.sky_area()[0].value, 25.0)
    assert np.allclose(box.area[0], 25.0 / 3600. / np.cos(10. /180 * np.pi))

    box = SRegion("BOX ICRS 90 10 5\" 5\"")
    assert box.N == 1

    assert np.allclose(box.centroid[0][0], 90, rtol=1.0e-3)
    assert np.allclose(box.centroid[0][1], 10, rtol=1.0e-3)
    assert np.allclose(box.sky_area()[0].value, 25.0 / 3600.)
    assert np.allclose(box.area[0], 25.0 / (3600.**2) / np.cos(10. /180 * np.pi))

    # BOX ICRS GEOCENTER string
    box = SRegion("BOX ICRS GEOCENTER 11.9 40.4 7.5 7.5")
    assert box.N == 1

    assert np.allclose(box.centroid[0][0], 11.9, rtol=1.0e-3)
    assert np.allclose(box.centroid[0][1], 40.4, rtol=1.0e-3)

    assert box.area[0] == 56.25


def test_stcs():
    """
    Test stripping values from STC-S specified strings
    """
    stcs = """Union ICRS ( Polygon 239.807341 -18.296691 239.803564 -18.300277 239.799786 -18.296691 239.803563
    -18.293105 Polygon 239.797826 -18.295944 239.794049 -18.299530 239.790272 -18.295944 239.794049
    -18.292358)"""

    sr = SRegion(stcs, verbose=False)

    assert len(sr.xy) == 2
    assert np.allclose(sr.area, 2.709e-5, rtol=0.01)


def test_whitespace():
    """ """
    pstr = "CIRCLE 5. 5. 1"
    for str_i in [pstr, pstr.replace(" ", "   "), "  " + pstr, pstr + "\n"]:
        sr = SRegion(str_i, ncircle=256)
        assert sr.N == 1
        assert np.allclose(sr.area, np.pi, rtol=1.0e-3)
        assert np.allclose(sr.centroid[0], 5.0, rtol=1.0e-3)

    pstr = (
        "POLYGON 0.000000 0.000000 0.000000 1.000000 "
        + "1.000000 1.000000 1.000000 0.000000"
    )

    for str_i in [pstr, pstr.replace(" ", "  "), "  " + pstr, pstr + "\n"]:
        sr = SRegion(str_i)
        assert sr.N == 1
        assert np.allclose(sr.area, 1.0)
        assert np.allclose(sr.centroid, 0.5)


def test_fromwcs():
    """
    Initialize from `astropy.wcs.WCS`
    """
    from astropy.io.fits import Header
    import astropy.wcs as pywcs

    # From WCS
    header_string = """
CRPIX1  =                  1.0                                                  
CRPIX2  =                  1.0                                                  
CRVAL1  =                 90.0                                                  
CRVAL2  =                  0.0                                                  
CD1_1   =             -0.00001                                                  
CD1_2   =                  0.0                                                  
CD2_1   =                  0.0                                                  
CD2_2   =              0.00001                                                  
NAXIS1  =                 1001                                                  
NAXIS2  =                 1001                                                  
CTYPE1  = 'RA---TAN'                                                            
CTYPE2  = 'DEC--TAN'                                                            
RADESYS = 'ICRS    '                                                            
EQUINOX =               2000.0                                                  
LATPOLE =                    0                                                  
LONPOLE =                180.0
    """

    head = Header.fromstring(header_string.replace("\n", ""))
    wcs = pywcs.WCS(head)
    sw = SRegion(wcs)
    pixel_area = np.abs(
        (head["NAXIS1"] - 1) * head["CD1_1"] * (head["NAXIS2"] - 1) * head["CD2_2"]
    )

    assert np.allclose(sw.area, pixel_area, atol=1.0e-5)

    assert np.allclose(sw.sky_area(unit=u.deg**2), pixel_area * u.deg**2)
    assert np.allclose(
        sw.sky_area(unit=u.arcmin**2), (pixel_area * u.deg**2).to(u.arcmin**2)
    )


def test_patch():
    """
    Test patch function
    """
    from matplotlib.patches import PathPatch
    circ = SRegion("CIRCLE 5. 5. 1", ncircle=256)
    _ = patch_from_polygon(circ.shapely[0], fc="k")


def test_draw_patch():
    """
    Draw in matplotlib axis
    """
    import matplotlib.pyplot as plt

    fig, ax = plt.subplots(1, 1)
    ax.plot([-10, 10], [-10, 10])

    circ = SRegion("CIRCLE 5. 5. 1", ncircle=256)

    circ.add_patch_to_axis(ax, fc="r", alpha=0.5)

    plt.close("all")


def test_convex_hull():

    x = np.array([0, 0, 1, 1, 0.5])
    y = np.array([0, 1, 1, 0, 0.5])

    srh = SRegion(np.array([x, y]).T, get_convex_hull=True)
    assert srh.xy[0].ndim == 2
    assert srh.area[0] == 1.0


def test_padding():

    x = np.array([0, 0, 1, 1])
    y = np.array([0, 1, 1, 0])

    sr = SRegion(np.array([x, y]).T, pad=2)
    assert sr.area[0] == 4.0

    sr = SRegion(np.array([x, y]).T)
    sr.pad(scale=2, in_place=True)
    assert sr.area[0] == 4.0

    sr = SRegion(np.array([x, y]).T, wrap=False)
    sr2 = sr.pad(scale=2., in_place=False)
    assert sr2.area[0] == 4.0
    assert np.allclose(sr2.centroid, 0.5)

    sr = SRegion(np.array([x, y]).T, wrap=False)
    sr2 = sr.pad(scale=2., center=[0,0], in_place=False)
    assert sr2.area[0] == 4.0
    assert np.allclose(sr2.centroid, 1.0)



"""
Utilities
"""

import numpy as np
from .sregion import SRegion


def concave_hull(
    points, smooth=0.08, sky=True, ref_density=250, scale_power=0.5, **kwargs
):
    """
    Compute the approximate concave hull of a set of points.

    Adapted from
    https://gist.github.com/jclosure/d93f39a6c7b1f24f8b92252800182889
    by GitHub user @jclosure

    Parameters
    ----------
    points : (N, 2) array
        Point list

    smooth : float
        Smoothing parameter

    sky : bool
        Treat ``points`` as sky coordinates in decimal degrees

    ref_density : float, None
        Reference sky surface density (N per square arcmin).
        If provided, scale the input ``smooth`` by the surface density of ``points``
        relative to ``ref_density``.

    scale_power : float
        Power on the relative surface density

    Returns
    -------
    output_hull : `~sregion.sregion.SRegion`
        Concave hull of the input ``points`` array

    scale_smooth : float
        Smoothing parameter used, perhaps rescaled

    """
    from shapely.ops import unary_union, polygonize
    from shapely import geometry
    from scipy.spatial import Delaunay, ConvexHull

    if points.shape[0] == 2:
        points = points.T

    if len(points) < 4:
        # When you have a triangle, there is no sense in computing an smooth
        # shape.
        return geometry.MultiPoint(list(points)).convex_hull

    def add_edge(edges, edge_points, coords, i, j):
        """Add a line between the i-th and j-th points, if not in the list already"""
        if (i, j) in edges or (j, i) in edges:
            # already added
            return
        edges.add((i, j))
        edge_points.append(coords[[i, j]])

    center = np.median(points, axis=0)
    cosd = np.array([np.cos(center[1] / 180 * np.pi), 1.0]) if sky else np.array([1, 1])

    if (ref_density is not None) & (sky):
        hull = ConvexHull(points)
        sr = SRegion(points[hull.vertices, :])
        scale_smooth = (
            smooth
            * (ref_density / (len(points) / sr.sky_area()[0].value)) ** scale_power
        )
    else:
        scale_smooth = smooth

    if sky:
        cosd *= 60

    coords = points - center
    coords *= cosd

    tri = Delaunay(coords)
    edges = set()
    edge_points = []

    # loop over triangles:
    # ia, ib, ic = indices of corner points of the triangle
    for ia, ib, ic in tri.simplices:
        pa = coords[ia]
        pb = coords[ib]
        pc = coords[ic]

        # Lengths of sides of triangle
        a = np.sqrt((pa[0] - pb[0]) ** 2 + (pa[1] - pb[1]) ** 2)
        b = np.sqrt((pb[0] - pc[0]) ** 2 + (pb[1] - pc[1]) ** 2)
        c = np.sqrt((pc[0] - pa[0]) ** 2 + (pc[1] - pa[1]) ** 2)

        # Semiperimeter of triangle
        s = (a + b + c) / 2.0

        # Area of triangle by Heron's formula
        area = np.sqrt(s * (s - a) * (s - b) * (s - c))
        circum_r = a * b * c / (4.0 * area)
        # radius = np.sqrt(area / np.pi)

        # Here's the radius filter.
        # print circum_r
        # if circum_r < 1.0/smooth:
        if circum_r < scale_smooth:
            add_edge(edges, edge_points, coords, ia, ib)
            add_edge(edges, edge_points, coords, ib, ic)
            add_edge(edges, edge_points, coords, ic, ia)

    edge_points = [p / cosd + center for p in edge_points]

    m = geometry.MultiLineString(edge_points)
    triangles = list(polygonize(m))
    output_hull = SRegion(unary_union(triangles))

    return output_hull, scale_smooth


def get_sky_footprint(ra, dec, make_figure=False, **kwargs):
    """
    Get the footprint of a list of sky coordinates

    Parameters
    ----------
    ra, dec : array-like
        Sky coordinates, decimal degrees

    make_figure : bool
        Make a diagnostic figure

    Returns
    -------
    catalog_hull : `~sregion.sregion.SRegion`
        Footprint

    fig : `~matplotlib.Figure`
        Figure object

    """
    import matplotlib.pyplot as plt

    points = np.array([ra, dec]).T
    catalog_hull, scale_smooth = concave_hull(points, **kwargs)

    if make_figure:
        fig, ax = plt.subplots(1, 1, figsize=(5, 5))
        ax.scatter(*points.T, color="0.5", alpha=0.05, marker=".", zorder=100)
        label = f"concave hull, area={catalog_hull.sky_area()[0]:.1f}"
        catalog_hull.add_patch_to_axis(ax, fc="tomato", ec="None", alpha=0.2)
        catalog_hull.add_patch_to_axis(
            ax, fc="None", ec="tomato", alpha=0.5, label=label
        )
        ax.set_xlim(*ax.get_xlim()[::-1])
    else:
        fig = None

    return catalog_hull, fig



def test_concave_hull():
    """
    Test concave hull function
    """
    import numpy as np

    from ..sregion import SRegion
    from ..utils import concave_hull

    # Demo concave polygon
    xy = np.array(
        [
            [-0.05340819, 0.01697288],
            [0.00633657, 0.08214911],
            [0.07965975, -0.01697297],
            [0.01991492, -0.00746813],
            [0.00090514, -0.03734045],
            [-0.05340819, -0.03734045],
        ]
    )

    nominal_area = 34.26
    xy[:, 0] += 10

    sr = SRegion(xy, wrap=True)

    # Random points within the test polygon
    np.random.seed(1)
    rnd = np.random.rand(2000, 2)

    mi, ma = xy.min(axis=0), xy.max(axis=0)
    si = ma - mi
    rnd = rnd * si + mi

    inside = sr.path[0].contains_points(rnd)
    rnd = rnd[inside, :]

    assert np.allclose(
        [area.value for area in sr.sky_area()], nominal_area, rtol=1.0e-2
    )

    # Test that get the same answer at different locations on the sky
    for ra in [45, 150]:
        for dec in [-80, -10, 0, 30, 50]:
            cosd = np.cos(dec / 180 * np.pi)

            xy_i = xy / np.array([cosd, 1.0]) + np.array([ra, dec])
            sr_i = SRegion(xy_i, wrap=True)

            assert np.allclose(
                [area.value for area in sr_i.sky_area()], nominal_area, rtol=1.0e-2
            )

            rnd_i = rnd / np.array([cosd, 1.0]) + np.array([ra, dec])

            hull_i, _alpha = concave_hull(
                rnd_i,
                alpha=0.08,
                sky=True,
                ref_density=250,
                scale_power=0.5,
            )

            print(ra, dec, hull_i.sky_area()[0], sr_i.sky_area()[0], nominal_area)

            assert hull_i.sky_area()[0] < sr_i.sky_area()[0]

            assert np.allclose(
                [area.value for area in hull_i.sky_area()], nominal_area, rtol=5.0e-2
            )



import numpy as np
from matplotlib.patches import PathPatch
from matplotlib.path import Path

import astropy.units as u
import astropy.wcs as pywcs
from astropy.coordinates import Angle

from shapely.geometry import Polygon

__all__ = ["SRegion", "patch_from_polygon"]


def _parse_circle(spli, ncircle=32):
    """
    Parse circle from split string

    Parameters
    ----------
    spli : list
        List of circle parameters ``['x', 'y', 'radius']``

    ncircle : int
        Number of points to sample around the diameter of the circle

    Returns
    -------
    poly_i : array-like
        2D xy array

    """
    if len(spli) != 3:
        msg = f"Input does not have three elements ({spli})"
        raise ValueError(msg)

    x0, y0 = np.asarray(spli[:2], dtype=float)
    cosd = np.cos(y0 / 180 * np.pi)

    r0 = spli[2]
    if r0.endswith('"'):
        scl = float(r0[:-1]) / 3600
    elif r0.endswith("'"):
        scl = float(r0[:-1]) / 60
    else:
        try:
            scl = float(r0)
        except ValueError:
            scl = 1.0

        cosd = 1.0

    _theta = np.linspace(0, 2 * np.pi, ncircle + 1)[:-1]
    _xc = np.cos(_theta)
    _yc = np.sin(_theta)

    poly_i = np.array([_xc / cosd * scl + x0, _yc * scl + y0]).T
    return poly_i


def _parse_box(spli):
    """
    Parse box from split string

    Parameters
    ----------
    spli : list
        List of box parameters ``['x', 'y', 'width', 'height']``

    Returns
    -------
    poly_i : array-like
        2D xy array
    """
    if len(spli) != 4:
        msg = f"Input does not have four elements ({spli})"
        raise ValueError(msg)

    # Extract center of box
    x0, y0 = np.asarray(spli[:2], dtype=float)
    dx, dy = spli[2], spli[3]

    # Parse optional sky coordinates    
    if dx.endswith('"'):
        cosd = np.cos(y0 / 180.0 * np.pi)
        half_width = float(dx[:-1]) / 3600 / cosd / 2
    elif dx.endswith("'"):
        cosd = np.cos(y0 / 180.0 * np.pi)
        half_width = float(dx[:-1]) / 60 / cosd / 2
    else:
        try:
            half_width = float(dx) / 2
        except ValueError:
            half_width = 1.0

    if dy.endswith('"'):
        cosd = np.cos(y0 / 180.0 * np.pi)
        half_height = float(dy[:-1]) / 3600 / 2
    elif dy.endswith("'"):
        cosd = np.cos(y0 / 180.0 * np.pi)
        half_height = float(dy[:-1]) / 60 / 2
    else:
        try:
            half_height = float(dy) / 2
        except ValueError:
            half_height = 1.0

    poly_i = np.array(
        [
            [x0 - half_width, y0 - half_height],
            [x0 + half_width, y0 - half_height],
            [x0 + half_width, y0 + half_height],
            [x0 - half_width, y0 + half_height],
        ]
    )
    return poly_i


def _wrap_xy(xy):
    """
    Wrap x coordinate as an angle

    Parameters
    ----------
    xy : (M,2) array
        Coordinates

    Returns
    -------
    wxy: (M,2) array
        Array with first column wrapped at 360 degrees

    """
    wxy = xy * 1.0
    ra = Angle(xy[:, 0] * u.deg).wrap_at(360 * u.deg).value
    wxy[:, 0] = ra
    return wxy


def _parse_sregion(sregion, ncircle=32, verbose=False, **kwargs):
    """
    Parse an S_REGION string with CIRCLE, POLYGON or BOX

    Parameters
    ----------
    sregion : str
        Region string

    Returns
    -------
    poly : list
        List of 2D xy arrays

    """
    if hasattr(sregion, "decode"):
        decoded = sregion.decode("utf-8").strip().upper()
    else:
        decoded = sregion.strip().upper()

    # Strip out prefix for some STS-C regions
    for strip_string in ["UNION", "ICRS", "FK5", "IMAGE", "WORLD"]:
        if (strip_string in decoded) & verbose:
            print("Strip {0} from {1}".format(strip_string, sregion))

        decoded = decoded.replace(strip_string, "")

    decoded = decoded.strip()
    if decoded.startswith("("):
        decoded = decoded[1:]

    if decoded.endswith(")"):
        decoded = decoded[:-1]

    decoded = decoded.strip()
    polyspl = (
        decoded.replace("POLYGON", "xxx").replace("CIRCLE", "xxx").replace("BOX", "xxx")
    )
    polyspl = polyspl.split("xxx")

    poly = []
    for pp in polyspl:
        if not pp:
            continue

        # strip out ()
        pp = pp.replace("(", "").replace(")", "")

        # ',' or ' ' delimiters
        if "," in pp:
            spl = pp.strip().split(",")
        else:
            spl = pp.strip().split()

        for ip, p in enumerate(spl):
            # Find index of first float
            try:
                _ = float(p)
                break
            except ValueError:
                continue

        if len(spl[ip:]) == 3:
            # Circle
            poly_i = _parse_circle(spl[ip:], ncircle=ncircle)
        elif len(spl[ip:]) == 4:
            # Box
            poly_i = _parse_box(spl[ip:])
        else:
            poly_i = np.asarray(spl[ip:], dtype=float).reshape((-1, 2))

        if len(poly_i) < 2:
            continue
        poly.append(poly_i)

    return poly


class SRegion(object):
    SREGION_PREFIX = "POLYGON ICRS"

    def __init__(self, inp, label=None, wrap=True, get_convex_hull=False, pad=None, **kwargs):
        """
        Helper class for parsing an S_REGION strings and general polygon
        tools

        Parameters
        ----------
        inp : str, (M,2) array, `~astropy.wcs.WCS`, `shapely.geometry.polygon.Polygon`

            Can be a "S_REGION" string, an array, or an `astropy.wcs.WCS`
            object from which the corners will be extracted with
            `~astropy.wcs.WCS.calc_footprint`.

        label : str
            Optional label attached to regions and patches

        wrap : bool
            Wrap first dimension as an angle between (0, 360) degrees

        get_convex_hull : bool
            Compute a `scipy.spatial.ConvexHull` from ``inp`` before generating
        """
        if isinstance(inp, str):
            self.xy = _parse_sregion(inp, **kwargs)

        elif hasattr(inp, "sum"):
            # NDarray
            sh = inp.shape
            if inp.ndim != 2:
                raise ValueError(f"Input shape {sh} is not (M,2)")
            else:
                if inp.shape[1] != 2:
                    if inp.shape[0] != 2:
                        raise ValueError(f"Input shape {sh} is not (M,2)")
                    else:
                        inp = inp.T

            if get_convex_hull:
                from scipy.spatial import ConvexHull
                inp = inp[ConvexHull(inp).vertices, :]

            self.xy = [inp]

        elif isinstance(inp, list):
            self.xy = inp

        elif isinstance(inp, pywcs.WCS):
            self.xy = [inp.calc_footprint()]

        elif hasattr(inp, "buffer"):
            # Shapely polygon
            if hasattr(inp, "geoms"):
                self.xy = [np.array(p.boundary.xy).T for p in inp.geoms]
            elif hasattr(inp, "__len__"):
                self.xy = [np.array(p.boundary.xy).T for p in inp]
            else:
                self.xy = [np.array(inp.boundary.xy).T]

        else:
            raise IOError("input must be ``str``, ``list``, or ``np.array``")

        if wrap:
            self.xy = [_wrap_xy(xy_i) for xy_i in self.xy]

        self.inp = inp
        self.ds9_properties = ""
        self.label = label
        self.is_wrapped = wrap

        if pad is not None:
            self.pad(scale=pad, in_place=True, **kwargs)

    @property
    def N(self):
        return len(self.xy)

    @property
    def centroid(self):
        return [np.mean(fp, axis=0) for fp in self.xy]

    def sky_buffer(self, buffer_deg):
        """
        Buffer polygons accounting for cos(dec)
        """

        new_xy = []
        for xy, c in zip(self.xy, self.centroid):
            cosd = np.array([1, np.cos(c[1] / 180 * np.pi)])
            p = Polygon((xy - c) * cosd).convex_hull.buffer(buffer_deg)
            xyp = np.array(p.boundary.xy).T
            new_xy.append((xyp / cosd) + c)

        self.xy = new_xy

    @property
    def path(self):
        """
        `~matplotlib.path.Path` object
        """
        return [Path(fp) for fp in self.xy]

    @property
    def shapely(self):
        """
        `~shapely.geometry.Polygon` object.
        """
        return [Polygon(fp).convex_hull for fp in self.xy]

    @property
    def geoms(self):
        """
        compatibility for shapely >= 1.8, 2.0
        """
        if hasattr(self.shapely, "geoms"):
            return self.shapely.geoms
        else:
            return self.shapely

    @property
    def area(self):
        """
        Area of shapely polygons
        """
        return [sh.area for sh in self.geoms]

    def sky_area(self, unit=u.arcmin**2):
        """
        Assuming coordinates provided are RA/Dec degrees, compute area
        """
        cosd = np.cos(self.centroid[0][1] / 180 * np.pi)
        return [(sh.area * cosd * u.deg**2).to(unit) for sh in self.geoms]

    def matplotlib_patch(self, **kwargs):
        """
        `~matplotlib.patches.PathPatch` object
        """
        if "label" not in kwargs:
            kwargs["label"] = self.label

        patches = [PathPatch(Path(xy), **kwargs) for xy in self.xy]

        return patches

    def descartes_patch(self, **kwargs):
        """
        `~descartes.PolygonPatch` object

        Deprecated because descartes doesn't seem to work with shapely>=2.0
        """
        from descartes import PolygonPatch

        if "label" not in kwargs:
            kwargs["label"] = self.label

        return [PolygonPatch(p, **kwargs) for p in self.geoms]

    def patch(self, **kwargs):
        """
        general patch object
        """
        try:
            return self.descartes_patch(**kwargs)
        except IndexError:
            return self.matplotlib_patch(**kwargs)
        except ImportError:
            return self.matplotlib_patch(**kwargs)

    def get_patch(self, **kwargs):
        """
        `~descartes.PolygonPatch` object

        ** Deprecated, use patch **
        """
        return self.patch(**kwargs)

    def add_patch_to_axis(self, ax, **kwargs):
        """
        Add patches to a plot axis

        Parameters
        ----------
        ax : matplotlib axis
            Axis in which do draw patches

        kwargs : dict
            Keyword arguments passed to `~sregion.SRegion.get_patch`

        """
        for patch in self.get_patch(**kwargs):
            ax.add_patch(patch)

    def union(self, shape=None, as_polygon=False):
        """
        Union of self and `shape` object.  If no `shape` provided, then
        return union of (optional) multiple components of self
        """
        if shape is None:
            un = self.shapely[0]
        else:
            un = shape

        for s in self.geoms:
            un = un.union(s)

        if as_polygon:
            return un
        else:
            return SRegion(un)

    def intersects(self, shape):
        """
        Union of self and `shape` object
        """
        test = False
        for s in self.geoms:
            test |= s.intersects(shape)

        return test

    @property
    def region(self):
        """
        Polygon string in DS9 region format
        """
        pstr = "polygon({0})"
        if hasattr(self, "ds9_properties"):
            tail = "{0}".format(self.ds9_properties)
        else:
            tail = ""

        if hasattr(self, "label"):
            if self.label is not None:
                tail += " text={xx}".replace("xx", self.label)

        if tail:
            tail = " # " + tail

        return [
            pstr.format(",".join([f"{c:.6f}" for c in fp.flatten()])) + tail
            for fp in self.xy
        ]

    @property
    def s_region(self):
        """
        Polygon as VO s_region
        """
        pstr = self.SREGION_PREFIX + " {0}"
        polys = [
            pstr.format(" ".join([f"{c:.6f}" for c in fp.flatten()])) for fp in self.xy
        ]
        return " ".join(polys)

    def polystr(self, precision=6):
        """
        String suitable for, e.g., SQL, e.g.,

        ((x1,y1),(x2,y2),...)
        """
        np.set_printoptions(precision=precision)

        fps = []
        for xy in self.xy:
            fmt = "({0:.pf},{1:.pf})".replace("p", f"{precision:0}")
            fpstr = [fmt.format(*row) for row in xy]
            fps.append("(" + ",".join(fpstr) + ")")

        return fps

    def pad(self, scale=1.5, center="centroid", sky=True, in_place=True, **kwargs):
        """
        Expand or shrink regions

        Parameters
        ----------
        scale : float
            Scale factor, > 1 expands, < 1 contracts

        sky : bool
            Account for cos(declination)

        in_place : bool
            Update `xy` attribute in-place

        Returns
        -------
        result : `sregion.SRegion`, None
            New region object if ``in_place = True``

        """
        xy_pad = []
        if center in ["centroid"]:
            x0 = self.centroid
        else:
            x0 = [center] * self.N

        for j in range(self.N):
            if sky:
                cosd = np.cos([x0[j][1]/180*np.pi, 0])
            else:
                cosd = np.ones(2)

            xy_pad.append(
                ((self.xy[j] - x0[j]) * cosd) * scale / cosd + x0[j]
            )

        if in_place:
            self.xy = xy_pad
            result = None
        else:
            result = SRegion(xy_pad, wrap=self.is_wrapped, label=self.label)

        return result


#### PatchPolygon
# descartes no longer works with shapely>=2.0
# this is similar do descartes and taken from
# https://codereview.stackexchange.com/questions/232294/plotting-shapely-polygon-in-matplotlib
# credit: Bruno Vermeulen and Graipher at stackexchange


class PatchPolygon:
    """ """

    def __init__(self, polygon, **kwargs):
        polygon_path = self.pathify(polygon)
        self._patch = PathPatch(polygon_path, **kwargs)

    @property
    def patch(self):
        """get a Patch object"""
        return self._patch

    @staticmethod
    def generate_codes(n):
        """The first command needs to be a "MOVETO" command,
        all following commands are "LINETO" commands.
        """
        return [Path.MOVETO] + [Path.LINETO] * (n - 1)

    def pathify(self, polygon):
        """Convert coordinates to path vertices. Objects produced by Shapely's
        analytic methods have the proper coordinate order, no need to sort.

        The codes will be all "LINETO" commands, except for "MOVETO"s at the
        beginning of each subpath
        """
        vertices = list(polygon.exterior.coords)
        codes = self.generate_codes(len(polygon.exterior.coords))

        for interior in polygon.interiors:
            vertices.extend(interior.coords)
            codes.extend(self.generate_codes(len(interior.coords)))

        return Path(vertices, codes)


def patch_from_polygon(polygon, **kwargs):
    """
    Generate the Path object from `PatchPolygon`

    Parameters
    ----------
    polygon : `shapely.geometry.Polygon`
        Input polygon object

    kwargs : dict
        Keywords to pass when generating the ``patch``

    Returns
    -------
    patch : `from matplotlib.patches.PathPatch`
        The patch that can be added to a `matplotlib` figure

    """
    obj = PatchPolygon(polygon, **kwargs)
    return obj._patch
